#ifndef CAMERA_H
#define CAMERA_H

#include "GLSL.h"
#include "glm/glm.hpp"

#define SPEED 10.0f  // Move SPEED units in world space per second
#define SPEED_FAST 4 // Move SPEED * SPEED_FAST units when moving faster

#define UP_VECTOR vec3(0, 1, 0)

#define EYE_SIZE 0.5f
#define EYE_START 8.5f
#define EYE_MIN 0.0f
#define EYE_MAX 180.0f

#define MAX_PITCH 80.0f

using glm::vec3;
using glm::mat4;

class Camera {
	vec3 eye;
	vec3 lookat;
	double curX, curY, lastY;
	double phi, theta;

public:
	Camera();

	void Init(GLFWwindow *window, float g_width, float g_height);

	// Camera motion
	void translate(GLFWwindow *window, vec3 lookat, float elapsed);
	void cursor_position_callback(GLFWwindow* window, double xpos, double ypos, float g_width, float g_height);

	void fixForScreenshot();

	vec3 getLookat() { return lookat; }
	vec3 getEye() { return eye; }

	mat4 getView();
	mat4 getViewLookat(vec3 lookat);

private:
	void clamp();
};

#endif