#include <stdio.h>
#include <string>
#include "LoadBMP.h"

BmpLoader::BmpLoader() :
  _data(NULL),
  _allocated_size(0)
{
}

BmpLoader::~BmpLoader()
{
  if (_data)
    delete[] _data;
}

unsigned char *BmpLoader::getBmpData(const std::string &path)
{
  FILE *fd;
  t_bmp_header header;
  t_bmp_img_header img_header;

  std::cout << "path : " << path << std::endl;
  if ((fd = fopen(path.c_str(), "rb")) == NULL)
    {
      std::cerr << "Cannot open file " << path << std::endl;
	  fclose(fd);
      return NULL;
    }
  if ((fread(&header, sizeof(header), 1, fd)) < 1)
    {
      std::cerr << "Cannot read file " << path << std::endl;
	  fclose(fd);
      return NULL;
    }
  if ((fread(&img_header, sizeof(img_header), 1, fd)) < 1)
    {
      std::cerr << "Cannot read file " << path << std::endl;
	  fclose(fd);
      return NULL;
    }

  _width = img_header.width;
  _height = img_header.height;
  _bpp = img_header.bpp / 8;
  unsigned int size = img_header.width * img_header.height * (img_header.bpp / 8);
  try
    {
      if (_allocated_size != size)
	{
	  delete[] _data;
	  _data = NULL;
	}
      if (!_data)
	{
	  std::cout << "allocated " << size << std::endl;
	  _data = new unsigned char[size];
	}
    }
  catch (std::exception& e)
    {
      std::cerr << "Error : " <<  e.what() << std::endl;
      fclose(fd);
      return NULL;
    }
  _allocated_size = size;
  fseek(fd, header.start_offset, SEEK_SET);
  if ((fread(_data, size, 1, fd)) < 1)
    {
      std::cerr << "Cannot read file data " << path << std::endl;
      fclose(fd);
      return NULL;
    }
  fclose(fd);
  return _data;
}

void setBmpData(int width, int height, int bpp, const std::string &path, void *pixels) {
	FILE *fd;
	t_bmp_header header;
	t_bmp_img_header img_header;
	
	img_header.img_header_size = sizeof(img_header);
	img_header.width = width;
	img_header.height = height;
	img_header.nb_plan = 1;
	img_header.bpp = bpp;
	img_header.compression = 0; //BI_RGB
	img_header.img_size = img_header.width * img_header.height * (img_header.bpp / 8);
	img_header.wppm = img_header.wppm = 10000;
	img_header.palette = 0;
	img_header.important_color = 0;
	
	header.magic = 0x4D42;
	header.start_offset = sizeof(header) + sizeof(img_header);
	header.bmp_size = header.start_offset + img_header.img_size;

	std::cout << "path : " << path << std::endl;
	if ((fd = fopen(path.c_str(), "wb")) == NULL) {
		std::cerr << "Cannot open file " << path << std::endl;
		fclose(fd);
	}
	if ((fwrite(&header, sizeof(header), 1, fd)) < 1)
	{
		std::cerr << "Cannot write file " << path << std::endl;
		fclose(fd);
	}
	if ((fwrite(&img_header, sizeof(img_header), 1, fd)) < 1)
	{
		std::cerr << "Cannot write file " << path << std::endl;
		fclose(fd);
	}
	if ((fwrite(pixels, img_header.img_size, 1, fd)) < 1)
	{
		std::cerr << "Cannot write file " << path << std::endl;
		fclose(fd);
	}

	fclose(fd);
}

void		BmpLoader::deleteData(void)
  {
    if (_data)
      delete[] _data;
    _data = NULL;
  }