#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <cassert>
#include <cmath>
#include <algorithm>
#include "Mesh.h"
#include "GLSL.h"
#include "tiny_obj_loader.h"
#include "glm/glm.hpp"
#include "glm/gtc/matrix_transform.hpp" //perspective, trans etc
#include "glm/gtc/type_ptr.hpp" //value_ptr
using namespace std;

Mesh::Mesh(const string &objFile) {
	loadObject(objFile);
}

void Mesh::loadObject(const string &objFile) {
	loadShapes(objFile);

	posBuf = _shapes[0].mesh.positions;
	glGenBuffers(1, &_posBufObj);
	glBindBuffer(GL_ARRAY_BUFFER, _posBufObj);
	glBufferData(GL_ARRAY_BUFFER, posBuf.size()*sizeof(float), &posBuf[0], GL_STATIC_DRAW);

	indBuf = _shapes[0].mesh.indices;
	glGenBuffers(1, &_indBufObj);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, _indBufObj);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, indBuf.size()*sizeof(unsigned int), &indBuf[0], GL_STATIC_DRAW);

	if (_shapes[0].mesh.normals.size()) {
		norBuf = _shapes[0].mesh.normals;
	}
	else {
		generateNormals();
	}
	glGenBuffers(1, &_norBufObj);
	glBindBuffer(GL_ARRAY_BUFFER, _norBufObj);
	glBufferData(GL_ARRAY_BUFFER, norBuf.size()*sizeof(float), &norBuf[0], GL_STATIC_DRAW);
}

void Mesh::loadObject(vector<float> positions, vector<float> normals, vector<unsigned int> indices) {
	posBuf = positions;
	glGenBuffers(1, &_posBufObj);
	glBindBuffer(GL_ARRAY_BUFFER, _posBufObj);
	glBufferData(GL_ARRAY_BUFFER, posBuf.size()*sizeof(float), &posBuf[0], GL_STATIC_DRAW);
	
	indBuf = indices;
	glGenBuffers(1, &_indBufObj);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, _indBufObj);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, indBuf.size()*sizeof(unsigned int), &indBuf[0], GL_STATIC_DRAW);

	if (normals.size()) {
		norBuf = normals;
	}
	else {
		generateNormals();
	}
	glGenBuffers(1, &_norBufObj);
	glBindBuffer(GL_ARRAY_BUFFER, _norBufObj);
	glBufferData(GL_ARRAY_BUFFER, norBuf.size()*sizeof(float), &norBuf[0], GL_STATIC_DRAW);
}

int Mesh::useMesh(GLuint aPos) {
	// Bind index array for drawing
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, indBufObj());

	// Enable and bind position array for drawing
	GLSL::enableVertexAttribArray(aPos);
	glBindBuffer(GL_ARRAY_BUFFER, posBufObj());
	glVertexAttribPointer(aPos, 3, GL_FLOAT, GL_FALSE, 0, 0);

	return indices();
}

int Mesh::useMesh(GLuint aPos, GLuint aNor) {
	// Bind index array for drawing
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, indBufObj());

	// Enable and bind position array for drawing
	GLSL::enableVertexAttribArray(aPos);
	glBindBuffer(GL_ARRAY_BUFFER, posBufObj());
	glVertexAttribPointer(aPos, 3, GL_FLOAT, GL_FALSE, 0, 0);
	
	// Enable and bind normal array for drawing
	GLSL::enableVertexAttribArray(aNor);
	glBindBuffer(GL_ARRAY_BUFFER, norBufObj());
	glVertexAttribPointer(aNor, 3, GL_FLOAT, GL_FALSE, 0, 0);

	return indices();
}

void Mesh::resize_obj() {
    float minX, minY, minZ;
    float maxX, maxY, maxZ;
    float scaleX, scaleY, scaleZ;
    float shiftX, shiftY, shiftZ;
    float epsilon = 0.001f;
	
	minX = minY = minZ = 1.1754E+38F;
	maxX = maxY = maxZ = -1.1754E+38F;

    //Go through all vertices to determine min and max of each dimension
    for (size_t i = 0; i < _shapes.size(); i++) {
        for (size_t v = 0; v < _shapes[i].mesh.positions.size() / 3; v++) {
            if(_shapes[i].mesh.positions[3*v+0] < minX) minX = _shapes[i].mesh.positions[3*v+0];
            if(_shapes[i].mesh.positions[3*v+0] > maxX) maxX = _shapes[i].mesh.positions[3*v+0];

            if(_shapes[i].mesh.positions[3*v+1] < minY) minY = _shapes[i].mesh.positions[3*v+1];
            if(_shapes[i].mesh.positions[3*v+1] > maxY) maxY = _shapes[i].mesh.positions[3*v+1];

            if(_shapes[i].mesh.positions[3*v+2] < minZ) minZ = _shapes[i].mesh.positions[3*v+2];
            if(_shapes[i].mesh.positions[3*v+2] > maxZ) maxZ = _shapes[i].mesh.positions[3*v+2];
        }
	}
    
   //From min and max compute necessary scale and shift for each dimension
   float maxExtent, xExtent, yExtent, zExtent;
   xExtent = maxX-minX;
   yExtent = maxY-minY;
   zExtent = maxZ-minZ;
   if (xExtent >= yExtent && xExtent >= zExtent) {
     maxExtent = xExtent;
   }
   if (yExtent >= xExtent && yExtent >= zExtent) {
     maxExtent = yExtent;
   }
   if (zExtent >= xExtent && zExtent >= yExtent) {
     maxExtent = zExtent;
   }
    scaleX = 2.0f /maxExtent;
    shiftX = minX + (xExtent / 2.0f);
    scaleY = 2.0f / maxExtent;
    shiftY = minY + (yExtent / 2.0f);
    scaleZ = 2.0f / maxExtent;
    shiftZ = minZ + (zExtent / 2.0f);

    //Go through all verticies shift and scale them
	for (size_t i = 0; i < _shapes.size(); i++) {
        for (size_t v = 0; v < _shapes[i].mesh.positions.size() / 3; v++) {
            _shapes[i].mesh.positions[3*v+0] = (_shapes[i].mesh.positions[3*v+0] - shiftX) * scaleX;
            assert(_shapes[i].mesh.positions[3*v+0] >= -1.0 - epsilon);
            assert(_shapes[i].mesh.positions[3*v+0] <= 1.0 + epsilon);
            _shapes[i].mesh.positions[3*v+1] = (_shapes[i].mesh.positions[3*v+1] - shiftY) * scaleY;
            assert(_shapes[i].mesh.positions[3*v+1] >= -1.0 - epsilon);
            assert(_shapes[i].mesh.positions[3*v+1] <= 1.0 + epsilon);
            _shapes[i].mesh.positions[3*v+2] = (_shapes[i].mesh.positions[3*v+2] - shiftZ) * scaleZ;
            assert(_shapes[i].mesh.positions[3*v+2] >= -1.0 - epsilon);
            assert(_shapes[i].mesh.positions[3*v+2] <= 1.0 + epsilon);
        }
    }
}

void Mesh::loadShapes(const string &objFile)
{
	string err = tinyobj::LoadObj(_shapes, _materials, objFile.c_str());
	if(!err.empty()) {
		cerr << err << endl;
	}
	resize_obj();
}

void Mesh::generateNormals() {
	int idx1, idx2, idx3;
	glm::vec3 v1, v2, v3, norm;
	//for every vertex initialize a normal to 0
	for (unsigned int j = 0; j < posBuf.size()/3; j++) {
		norBuf.push_back(0);
		norBuf.push_back(0);
		norBuf.push_back(0);
	}

	for (unsigned int i = 0; i < indBuf.size()/3; i++) {
      idx1 = indBuf[3*i+0];
      idx2 = indBuf[3*i+1];
      idx3 = indBuf[3*i+2];
      v1 = glm::vec3(posBuf[3*idx1 +0],posBuf[3*idx1 +1], posBuf[3*idx1 +2]); 
      v2 = glm::vec3(posBuf[3*idx2 +0],posBuf[3*idx2 +1], posBuf[3*idx2 +2]); 
      v3 = glm::vec3(posBuf[3*idx3 +0],posBuf[3*idx3 +1], posBuf[3*idx3 +2]); 
     
	  norm = glm::cross(v2 - v1, v3 - v1);

      //This is not correct, it sets the normal as the vertex value but
      //shows access pattern 
      norBuf[3*idx1+0] += norm.x;
      norBuf[3*idx1+1] += norm.y;
      norBuf[3*idx1+2] += norm.z;

      norBuf[3*idx2+0] += norm.x;
      norBuf[3*idx2+1] += norm.y;
      norBuf[3*idx2+2] += norm.z;

      norBuf[3*idx3+0] += norm.x;
      norBuf[3*idx3+1] += norm.y;
      norBuf[3*idx3+2] += norm.z;
   }

	for (unsigned int j = 0; j < posBuf.size()/3; j++) {
      norm = glm::normalize(glm::vec3(norBuf[3*j+0], norBuf[3*j+1], norBuf[3*j+2]));
      
	  norBuf[3*j+0] = norm.x;
      norBuf[3*j+1] = norm.y;
      norBuf[3*j+2] = norm.z;
	}
}