//
// seriously modified ZJ Wood January 2015 - conversion to glfw
// inclusion of matrix stack Feb. 2015
// original from Shinjiro Sueda
// October, 2014
//

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <iostream>
#include <cassert>
#include "GLSL.h"
#include "tiny_obj_loader.h"
#include "Mesh.h"
#include "shader_program.h"
#include "glm/glm.hpp"
#include "glm/gtc/matrix_transform.hpp" //perspective, trans etc
#include "glm/gtc/type_ptr.hpp" //value_ptr
#include "RenderingHelper.h"
#include "LoadBMP.h"
#include "instance_types.h"
#include "Camera.h"

#define PI 3.1415926f

#define MAX_FPS 59
#define MAX_BLUR 1
#define ZNEAR 0.1f
#define ZFAR 300.0f

#define CANDLE_WICK 0.1f
#define CANDLE_SIZE 0.5f

enum GBUFFER_TEXTURE_TYPE {
	GBUFFER_TEXTURE_TYPE_POSITION,
	GBUFFER_TEXTURE_TYPE_NORMAL,
	GBUFFER_TEXTURE_TYPE_DIFFUSE,
	GBUFFER_NUM_TEXTURES
};

#define TEXTURE_BLUR_H 3
#define TEXTURE_BLUR_V 4
#define TEXTURE_AMB    5
#define TEXTURE_FINAL  6
#define TEXTURE_FIRE   7

enum DRAW_MODES {
	DRAW_FULL,
	DRAW_POSITION,
	DRAW_NORMAL,
	DRAW_DIFFUSE,
	DRAW_DEPTH,
	DRAW_AMB,
	DRAW_NUM_MODES
};

enum HELD_ITEMS {
	HELD_NONE,
	HELD_MATCH,
	//HELD_FLASHLIGHT,
	NUM_HELD_ITEMS
};

using namespace std;

GLFWwindow* window;
Mesh sphere, cube, plane, fire, bunny, candle, pedestal, tree, shrine, lamp;
Shader_Program geomPass, lightPass, dirLightPass, nullPass, bloomPass, skyDraw, fireDraw, texDraw, blurVert, blurHoriz, ambO;
RenderingHelper ModelTrans;

GLuint frameBufObj = 0;
GLuint Tex[GBUFFER_NUM_TEXTURES];
GLuint depthTex = 0, finalTex = 0, blurHTex = 0, blurVTex = 0, ambTex = 0, fireTex = 0;

Camera cam;

float g_width, g_height;
float angle;
vector<Fire> flames;
vector<Tree> trees;
PointLight source, moon;

int drawMode;
int blur, bloom, amb, dof, transfix, bright;
Fire match;
int matchpos, torch1pos, torch2pos;
int held;

unsigned char *BMP;
GLuint skyboxTex = 0;

float ambO_dist = 5.0f, ambO_rad = 20.0f, blur_rad = 2.0f;

// Angle between -90 and +90 degrees
float rand_angle() {
	return -90.0f + (float)rand() / RAND_MAX * 180.0f;
}

// Normalized vector in a unit sphere
vec3 rand_vector() {
	return glm::normalize(vec3(-1.0f + (float)rand() / RAND_MAX * 2.0f,
		-1.0f + (float)rand() / RAND_MAX * 2.0f,
		-1.0f + (float)rand() / RAND_MAX * 2.0f));
}

// Pass along the callback to the FPS camera
void cursor_position_callback(GLFWwindow* window, double xpos, double ypos) {
	cam.cursor_position_callback(window, xpos, ypos, g_width, g_height);	
}

/* 
Left Mouse Button   - Light/unlight the held match
Right Mouse Button  - Change the color of the lit match
Middle Mouse Button - Place a tree at the center of screen (imaginary crosshairs location)
*/
void mouse_button_callback(GLFWwindow *window, int button, int action, int mods) {
	if (button == GLFW_MOUSE_BUTTON_3 && action == GLFW_PRESS) {
		vector<float> pixels(1 * 1 * 3, 0);
		glBindFramebuffer(GL_READ_FRAMEBUFFER, frameBufObj);
		glReadBuffer(GL_COLOR_ATTACHMENT0 + GBUFFER_TEXTURE_TYPE_POSITION);
		glReadPixels((int)g_width / 2, (int)g_height / 2, 1, 1, GL_RGB, GL_FLOAT, &pixels[0]);
		vec4 pos(pixels[0], pixels[1], pixels[2], 1.0f);
		mat4 View = cam.getView(); 
		pos = glm::inverse(View) * pos;
		printf("(%f, %f, %f)\n", pos.x, pos.y, pos.z);

		/* PLACE CANDLE
		Fire temp;
		temp.source = source;
		temp.source.Position.y = 2 * ALTAR_HEIGHT + CANDLE_WICK + (float)rand() / RAND_MAX;
		temp.source.Position.x = pos.x;
		temp.source.Position.z = pos.z;
		temp.source.Color = vec3(1.0, 1.0, 1.0) + rand_vector();
		temp.outsideColor = vec3(1.0, 1.0, 1.0) + rand_vector();
		temp.source.Position -= altarPos;
		temp.lit = 1;

		for (int j = 0; j < FLAME_PARTICLE_COUNT; j++) {
			temp.life[j] = (float)j / (FLAME_PARTICLE_COUNT - 1);
			temp.sway[j] = rand_vector() * MAX_FIRE_SWAY;
		}
		
		flames.push_back(temp);*/

		/* PLACE TREE */
		Tree temp;
		temp.Position = vec3(pos);
		temp.rotation = 2 * rand_angle();
		temp.size = (2 + (float)rand() / RAND_MAX) * ALTAR_SIZE;
		trees.push_back(temp);
	}

	// Light/unlight the held match
	if (button == GLFW_MOUSE_BUTTON_1 && action == GLFW_PRESS) {
		if (held == HELD_MATCH) {
			match.lit = !match.lit;
			if (match.lit) {
				matchpos = flames.size();
				flames.push_back(match);
			}
			else
				flames.pop_back();
		}
	}

	// Change the color of the lit match
	if (button == GLFW_MOUSE_BUTTON_2 && action == GLFW_PRESS) {
		match.source.Color = vec3(1.0, 1.0, 1.0) + rand_vector();
		match.outsideColor = vec3(1.0, 1.0, 1.0) + rand_vector();
	}
}

/*
Print Screen - Take a screenshot (saved in folder /screenshot/)
Enter        - Cycle the draw mode
B            - Cycle the blur amount
M            - Toggle bloom (and blur)
O            - Toggle ambient occlusion
P            - Toggle depth of field (and blur)
L            - Toggle day/night lighting
F            - Take out or put away held match
UP/DOWN      - Raise/Lower ambient occlusion distance step
LEFT/RIGHT   - Raise/Lower ambient occlusion sample radius
-/+          - Rasie/Lower blur radius
*/
void key_callback(GLFWwindow *window, int key, int scancode, int action, int mods) {
	if (key == GLFW_KEY_PRINT_SCREEN && action == GLFW_PRESS) {
		unsigned char *pixels = new unsigned char[(int)(3 * g_width * g_height)];

		glBindFramebuffer(GL_READ_FRAMEBUFFER, 0);
		glReadPixels(0, 0, g_width, g_height, GL_BGR, GL_UNSIGNED_BYTE, pixels);
		setBmpData(g_width, g_height, 24, "screenshots/screenshot" + to_string(time(NULL)) + ".bmp", pixels); // Write screen to bitmap
		glBindFramebuffer(GL_READ_FRAMEBUFFER, frameBufObj);
	}

	// Change Draw Mode
	if (key == GLFW_KEY_ENTER && action == GLFW_PRESS) {
		drawMode = (drawMode + 1) % DRAW_NUM_MODES;
	}

	// Change Blur amount
	if (key == GLFW_KEY_B && action == GLFW_PRESS) {
		blur = (blur + 1) % (MAX_BLUR + 1);
		printf("BLUR %d\n", blur);
	}

	// Toggle Bloom (and Blur)
	if (key == GLFW_KEY_M && action == GLFW_PRESS) {
		bloom = !bloom;
		blur = (bloom) ? MAX_BLUR : 0; 
		printf("BLOOM %s\n", (bloom) ? "ON" : "OFF");
	}

	// Toggle Ambient Occlusion
	if (key == GLFW_KEY_O && action == GLFW_PRESS) {
		amb = !amb;
		printf("AMB %s\n", (amb) ? "ON" : "OFF");
	}

	// Toggle Depth of Field (and Blur)
	if (key == GLFW_KEY_P && action == GLFW_PRESS) {
		blur = dof = !dof;
		printf("DOF %s\n", (dof) ? "ON" : "OFF");
	}

	// Toggle Day/Night Lighting
	if (key == GLFW_KEY_L && action == GLFW_PRESS) {
		bright = !bright;
		if (bright)
			moon.Color = vec3(0.8f, 0.81f, 0.65f);
		else
			moon.Color = vec3(0.1f, 0.11f, 0.35f);
	}

	// Toggle Held Item
	if (key == GLFW_KEY_F && action == GLFW_PRESS) {
		held = (held + 1) % NUM_HELD_ITEMS;
		if (match.lit) {
			match.lit = 0;
			flames.pop_back();
		}
	}

	// Modify Distance Stepping of Ambient Occlusion Pass
	if (key == GLFW_KEY_UP && (action == GLFW_PRESS || action == GLFW_REPEAT)) {
		ambO_dist += 0.1f;
		printf("ambO_dist %f\n", ambO_dist);
	}
	if (key == GLFW_KEY_DOWN && (action == GLFW_PRESS || action == GLFW_REPEAT)) {
		ambO_dist -= 0.1f;
		printf("ambO_dist %f\n", ambO_dist);
	}

	// Modify Sampling Radius of Ambient Occlusion Pass
	if (key == GLFW_KEY_LEFT && (action == GLFW_PRESS || action == GLFW_REPEAT)) {
		ambO_rad -= 0.5f;
		printf("ambO_rad %f\n", ambO_rad);
	}
	if (key == GLFW_KEY_RIGHT && (action == GLFW_PRESS || action == GLFW_REPEAT)) {
		ambO_rad += 0.5f;
		printf("ambO_rad %f\n", ambO_rad);
	}

	// Modify Blur Radius of all Blur Passes
	if (key == GLFW_KEY_MINUS && (action == GLFW_PRESS || action == GLFW_REPEAT)) {
		blur_rad -= 0.25f;
		printf("blur_rad %f\n", blur_rad);
	}
	if (key == GLFW_KEY_EQUAL && (action == GLFW_PRESS || action == GLFW_REPEAT)) {
		blur_rad += 0.25f;
		printf("blur_rad %f\n", blur_rad);
	}
}

// Outer radial extent of light based on intensity, color, and attenuation
// Solution to quadratic equation for the best a 256-bit color buffer can represent
// http://ogldev.atspace.co.uk
float calcLightSphere(const PointLight &light) {
	return (-light.Attenuation.y + sqrtf(light.Attenuation.y * light.Attenuation.y - 
		4 * light.Attenuation.z * (light.Attenuation.x - 256 * std::max(std::max(light.Color.x, light.Color.y), light.Color.z) * light.DifIntensity))) / 
		(2 * light.Attenuation.z);
}

// Read and create VBOs for all meshes
void loadMeshes() {
	// Load a plane "Mesh"
	const float planePosBuf[] = {
		0, 0, 0,
		0, 1, 0,
		1, 1, 0,
		1, 0, 0
	};

	const float planeNorBuf[] = {
		0, 0, 1,
		0, 0, 1,
		0, 0, 1,
		0, 0, 1
	};

	const unsigned int planeIndBuf[] = {
		2, 1, 0,
		3, 2, 0
	};

	vector<float> pos(planePosBuf, planePosBuf + sizeof(planePosBuf) / sizeof(float));
	vector<float> nor(planeNorBuf, planeNorBuf + sizeof(planeNorBuf) / sizeof(float));
	vector<unsigned int> ind(planeIndBuf, planeIndBuf + sizeof(planeIndBuf) / sizeof(unsigned int));

	plane.loadObject(pos, nor, ind);

	// Load other Meshes
	sphere.loadObject  ("meshes/sphere.obj");
	cube.loadObject    ("meshes/cube.obj");
	bunny.loadObject   ("meshes/bunny.obj");
	candle.loadObject  ("meshes/candle.obj");
	pedestal.loadObject("meshes/pedestal.obj");
	tree.loadObject    ("meshes/tree.obj");
	fire.loadObject    ("meshes/fire.obj");
	shrine.loadObject  ("meshes/shrine.obj");
	lamp.loadObject    ("meshes/t_oribe.obj");
}

// Compile and Link a Shader
// Query it for variable names
void initShader(Shader_Program &prog, const string &vert, const string &frag) {
	prog.setVertexShader(vert);
	prog.setFragShader(frag);
	prog.loadProgram();
	prog.queryAttributes();
	prog.queryUniforms();
}

// Compile and link all shaders into shader programs
void loadShaders() {
	initShader(geomPass,     "shaders/geomVert.glsl",   "shaders/geomFrag.glsl");
	initShader(lightPass,    "shaders/lightVert.glsl",  "shaders/ptLightFrag.glsl");
	initShader(dirLightPass, "shaders/lightVert.glsl",  "shaders/dirLightFrag.glsl");
	initShader(nullPass,     "shaders/nullVert.glsl",   "shaders/nullFrag.glsl");
	initShader(texDraw,      "shaders/texVert.glsl",    "shaders/texFrag.glsl");
	initShader(blurHoriz,    "shaders/nullVert.glsl",   "shaders/blurHorizFrag.glsl");
	initShader(blurVert,     "shaders/nullVert.glsl",   "shaders/blurVertFrag.glsl");
	initShader(bloomPass,    "shaders/texVert.glsl",    "shaders/bloomFrag.glsl");
	initShader(ambO,         "shaders/texVert.glsl",    "shaders/aoFrag.glsl");
	initShader(fireDraw,     "shaders/fireVert.glsl",   "shaders/fireFrag.glsl");
	initShader(skyDraw,      "shaders/skyboxVert.glsl", "shaders/skyboxFrag.glsl");
}

// Initialize the light sources
void setupLights() {
	source.Color = vec3(1.0f, 0.8f, 0.35f);
	source.AmbIntensity = 0.2f;
	source.DifIntensity = 0.6f;
	source.SpeIntensity = 0.1f;
	source.SpePower = 50.0f;
	source.Position = vec3(0.0f, EYE_START, 0.0f);
	source.Attenuation = vec3(5.0f, 1.0f, 0.5f);

	moon.Color = vec3(0.1f, 0.11f, 0.35f);
	moon.AmbIntensity = 0.5f;
	moon.DifIntensity = 0.2f;
	moon.SpeIntensity = 0.07f;
	moon.SpePower = 5.0f;
	moon.Position = vec3(1.0f, 1.0f, 1.0f);
}

// Initialize the Framebuffer with its textures
void setupFramebuffer() {
	glGenFramebuffers(1, &frameBufObj);
	glBindFramebuffer(GL_DRAW_FRAMEBUFFER, frameBufObj);

	// Create Space for Textures
	glGenTextures(GBUFFER_NUM_TEXTURES, Tex);
	glGenTextures(1, &depthTex);
	glGenTextures(1, &finalTex);
	glGenTextures(1, &blurHTex);
	glGenTextures(1, &blurVTex);
	glGenTextures(1, &ambTex);
	glGenTextures(1, &fireTex);

	// G-Buffer (Position, Normals, Diffuse Color)
	GLenum buffers[GBUFFER_NUM_TEXTURES];
	for (int i = 0; i < GBUFFER_NUM_TEXTURES; i++) {
		glBindTexture(GL_TEXTURE_2D, Tex[i]);
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB32F, (int)g_width, (int)g_height, 0, GL_RGB, GL_FLOAT, NULL);
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
		glFramebufferTexture2D(GL_DRAW_FRAMEBUFFER, GL_COLOR_ATTACHMENT0 + i, GL_TEXTURE_2D, Tex[i], 0);
	
		buffers[i] = GL_COLOR_ATTACHMENT0 + i;
	}

	// Horizontal Blur
	glBindTexture(GL_TEXTURE_2D, blurHTex);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB32F, (int)g_width, (int)g_height, 0, GL_RGB, GL_FLOAT, NULL);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glFramebufferTexture2D(GL_DRAW_FRAMEBUFFER, GL_COLOR_ATTACHMENT0 + TEXTURE_BLUR_H, GL_TEXTURE_2D, blurHTex, 0);
	
	// Vertical Blur
	glBindTexture(GL_TEXTURE_2D, blurVTex);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB32F, (int)g_width, (int)g_height, 0, GL_RGB, GL_FLOAT, NULL);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glFramebufferTexture2D(GL_DRAW_FRAMEBUFFER, GL_COLOR_ATTACHMENT0 + TEXTURE_BLUR_V, GL_TEXTURE_2D, blurVTex, 0);

	// Ambient Occlusion
	glBindTexture(GL_TEXTURE_2D, ambTex);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB32F, (int)g_width, (int)g_height, 0, GL_RGB, GL_FLOAT, NULL);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glFramebufferTexture2D(GL_DRAW_FRAMEBUFFER, GL_COLOR_ATTACHMENT0 + TEXTURE_AMB, GL_TEXTURE_2D, ambTex, 0);

	// Flames
	glBindTexture(GL_TEXTURE_2D, fireTex);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA32F, (int)g_width, (int)g_height, 0, GL_RGB, GL_FLOAT, NULL);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glFramebufferTexture2D(GL_DRAW_FRAMEBUFFER, GL_COLOR_ATTACHMENT0 + TEXTURE_FIRE, GL_TEXTURE_2D, fireTex, 0);

	// Depth
	glBindTexture(GL_TEXTURE_2D, depthTex);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_DEPTH24_STENCIL8, (int)g_width, (int)g_height, 0, GL_DEPTH_STENCIL, GL_FLOAT_32_UNSIGNED_INT_24_8_REV, NULL);
	glFramebufferTexture2D(GL_DRAW_FRAMEBUFFER, GL_DEPTH_STENCIL_ATTACHMENT, GL_TEXTURE_2D, depthTex, 0);

	// Output to Screen
	glBindTexture(GL_TEXTURE_2D, finalTex);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB32F, (int)g_width, (int)g_height, 0, GL_RGB, GL_FLOAT, NULL);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glFramebufferTexture2D(GL_DRAW_FRAMEBUFFER, GL_COLOR_ATTACHMENT0 + TEXTURE_FINAL, GL_TEXTURE_2D, finalTex, 0);

	assert(glCheckFramebufferStatus(GL_FRAMEBUFFER) == GL_FRAMEBUFFER_COMPLETE);
	glBindFramebuffer(GL_DRAW_FRAMEBUFFER, 0);
}

// Setup basic OpenGL parameters
void initGL()
{
	// Set the background color
	glClearColor(0.0f, 0.0f, 0.0f, 1.0f);

	// Enable Z-buffer test
	glDisable(GL_BLEND);
    glEnable(GL_TEXTURE_2D);

	setupFramebuffer();

	GLSL::checkVersion();
	assert(glGetError() == GL_NO_ERROR);

   //initialize the modeltrans matrix stack
   ModelTrans.useModelViewMatrix();
   ModelTrans.loadIdentity();
}

// Create the Skybox CubeMap
void loadSkybox() {
	BmpLoader skybox;

    glGenTextures(1, &skyboxTex);
	glBindTexture(GL_TEXTURE_CUBE_MAP, skyboxTex);

	BMP = skybox.getBmpData("bitmaps/nnksky01_back.bmp");
	glTexImage2D(GL_TEXTURE_CUBE_MAP_NEGATIVE_Z, 0, GL_RGB, skybox.getWidth(), skybox.getHeight(), 0, GL_BGR, GL_UNSIGNED_BYTE, BMP);
	skybox.deleteData();

	BMP = skybox.getBmpData("bitmaps/nnksky01_front.bmp");
	glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_Z, 0, GL_RGB, skybox.getWidth(), skybox.getHeight(), 0, GL_BGR, GL_UNSIGNED_BYTE, BMP);
	skybox.deleteData();

	BMP = skybox.getBmpData("bitmaps/nnksky01_left.bmp");
	glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_X, 0, GL_RGB, skybox.getWidth(), skybox.getHeight(), 0, GL_BGR, GL_UNSIGNED_BYTE, BMP);
	skybox.deleteData();

	BMP = skybox.getBmpData("bitmaps/nnksky01_right.bmp");
	glTexImage2D(GL_TEXTURE_CUBE_MAP_NEGATIVE_X, 0, GL_RGB, skybox.getWidth(), skybox.getHeight(), 0, GL_BGR, GL_UNSIGNED_BYTE, BMP);
	skybox.deleteData();

	BMP = skybox.getBmpData("bitmaps/nnksky01_top.bmp");
	glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_Y, 0, GL_RGB, skybox.getWidth(), skybox.getHeight(), 0, GL_BGR, GL_UNSIGNED_BYTE, BMP);
	glTexImage2D(GL_TEXTURE_CUBE_MAP_NEGATIVE_Y, 0, GL_RGB, skybox.getWidth(), skybox.getHeight(), 0, GL_BGR, GL_UNSIGNED_BYTE, BMP);
	skybox.deleteData();

	glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE);
}

void setupGLFW() {
	// Initialise GLFW
	if( !glfwInit() ) {
		fprintf( stderr, "Failed to initialize GLFW\n" );
		exit(-1);
	}

	glfwWindowHint(GLFW_SAMPLES, 4);
	glfwWindowHint(GLFW_RESIZABLE,GL_FALSE);
	glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 2);
	glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 1);


	// Open a window and create its OpenGL context
	window = glfwCreateWindow((int)g_width, (int)g_height, "Program 4", NULL, NULL);
	if( window == NULL ){
		fprintf( stderr, "Failed to open GLFW window. If you have an Intel GPU, they are not 3.3 compatible. Try the 2.1 version of the tutorials.\n" );
		glfwTerminate();
		exit(-1);
	}
	glfwMakeContextCurrent(window);

	// Ensure we can capture the escape key being pressed below
	glfwSetInputMode(window, GLFW_STICKY_KEYS, GL_TRUE);
	glfwSetKeyCallback(window, key_callback);
	glfwSetMouseButtonCallback(window, mouse_button_callback);

	// FPS Camera
	glfwSetInputMode(window, GLFW_CURSOR, GLFW_CURSOR_DISABLED);
	glfwSetCursorPosCallback(window, cursor_position_callback);
	cam.Init(window, g_width, g_height);
}

// Generate the placement of objects in the scene
void generate_scene() {
	srand((unsigned int)time(NULL));

	// Pregenerated positions of candles and lamps
	vec3 firepos[] = {
		vec3(-1.08050776, 4.69767427, -5.00000000),
		vec3(-2.52268505, 4.42975235, -4.58080292),
		vec3(-3.48236346, 5.08794498, -2.93612671),
		vec3(-4.79690170, 4.83247471, -2.55805969),
		vec3(-4.49243689, 4.88142633, -5.04150391),
		vec3(-3.29216337, 4.47043371, -5.10211182),
		vec3(1.14799416, 4.47385168, -5.00471497),
		vec3(3.00798893, 4.51404476, -3.83312225),
		vec3(2.46034217, 4.30514526, -4.18035889),
		vec3(4.49059486, 4.34903097, -4.84365845),
		vec3(5.06255484, 4.22131109, -3.25307465),
		vec3(3.88098454, 4.18673372, -2.05783081),
		vec3(4.85369301, 5.05828094, -1.22272491),
		vec3(4.30772972, 4.34585714, -3.56278992),
		vec3(4.53285456, 4.71625996, 0.659172058),
		vec3(2.88790178, 4.13384485, 2.22352600),
		vec3(4.23113823, 4.85756111, 1.66361237),
		vec3(5.16788960, 4.69465303, 3.40256500),
		vec3(4.00893402, 4.16955137, 4.03799438),
		vec3(5.07854700, 4.85572958, 5.00124359),
		vec3(3.23217821, 4.38461542, 3.83563232),
		vec3(3.71375966, 4.51288462, 5.00019073),
		vec3(2.10473347, 4.37854242, 4.39825439),
		vec3(0.460689306, 4.66880369, 4.30051422),
		vec3(-0.385008574, 4.57083941, 4.84259033),
		vec3(-1.79930055, 4.51477718, 3.57837677),
		vec3(-1.29401088, 4.24035454, 3.75373077),
		vec3(-2.77007198, 5.07769108, 5.03321838),
		vec3(-3.17339134, 4.89992046, 1.89123535),
		vec3(-2.71827745, 4.92674637, 3.35131073),
		vec3(-4.69436216, 4.35250998, 2.10157776),
		vec3(-4.74371195, 4.25488138, 0.570869446),
		vec3(-3.89073896, 4.78474379, 3.33368683),
		vec3(-4.81303215, 4.92930984, 4.99268341),
		vec3(-5.39041805, 4.67612839, -0.663879395),
		vec3(3.74995279, 5.01973629, -4.62178040),
		vec3(-3.49655628, 4.85954475, -3.80353546),
		vec3(-4.15993738, 4.51468563, 4.30868530),
		vec3(-0.327732563, 4.77043056, 2.70856476),
		vec3(1.68247509, 5.00163889, 3.03553772),
		vec3(2 * ALTAR_SIZE, 2.5f * ALTAR_HEIGHT, -2 * ALTAR_SIZE),
		vec3(-2 * ALTAR_SIZE, 2.5f * ALTAR_HEIGHT, -2 * ALTAR_SIZE),
	};

	// Initialize the flames array with random colors
	Fire tempFire;
	for (int i = 0; i < FLAME_COUNT; i++) {
		tempFire.source = source;
		if (i < sizeof(firepos) / sizeof(vec3))
			tempFire.source.Position = firepos[i];
		else 
			tempFire.source.Position = vec3(
			ALTAR_SIZE * (-1 + 2 * (float)rand() / RAND_MAX), 
			2 * ALTAR_HEIGHT + CANDLE_WICK + (float)rand() / RAND_MAX, 
			ALTAR_SIZE * (-1 + 2 * (float)rand() / RAND_MAX)
		);
		tempFire.source.Color = vec3(1.0, 1.0, 1.0) + rand_vector();
		tempFire.outsideColor = vec3(1.0, 1.0, 1.0) + rand_vector();
		tempFire.lit = rand() % 2;

		for (int j = 0; j < FLAME_PARTICLE_COUNT; j++) {
			tempFire.life[j] = (float)j / (FLAME_PARTICLE_COUNT - 1);
			tempFire.sway[j] = rand_vector() * 0.01f;
		}
		
		flames.push_back(tempFire);
	}

	// Setup match
	match.source = source;
	match.source.Color = vec3(0.8, 0.8, 0.15);
	match.outsideColor = vec3(0.8, 0.3, 0.25);
	match.lit = 0;
	for (int j = 0; j < FLAME_PARTICLE_COUNT; j++) {
		match.life[j] = (float)j / (FLAME_PARTICLE_COUNT - 1);
		match.sway[j] = rand_vector() * 0.01f;
	}

	// Setup lamps (last 2)
	torch1pos = FLAME_COUNT - 2;
	torch2pos = FLAME_COUNT - 1;
	flames[torch1pos].source.Attenuation = flames[torch2pos].source.Attenuation = vec3(2.0f, 0.5f, 0.2f);
	flames[torch1pos].lit = flames[torch2pos].lit = 0;


	// Pregenerated tree positions
	vec3 treepos[] = {
		vec3(-17.7726116, 0.0, 37.2579803),
		vec3(28.1853867, 0.0, 37.6184006),
		vec3(58.7162704, 0.0, 32.0372429),
		vec3(78.0750732, 0.0, 48.2984390),
		vec3(46.5938148, 0.0, 72.8217926),
		vec3(73.5965576, 0.0, 88.8579636),
		vec3(57.0857239, 0.0, 114.056557),
		vec3(28.7313881, 0.0, 146.964386),
		vec3(-16.3121262, 0.0, 154.252304),
		vec3(-41.4609070, 0.0, 105.080498),
		vec3(-64.7306595, 0.0, 153.121384),
		vec3(-80.3498230, 0.0, 84.9375534),
		vec3(15.6554356, 0.0, 181.676651),
		vec3(58.8853645, 0.0, 176.439651),
		vec3(100.672478, 0.0, 125.760262),
		vec3(108.060150, 0.0, 71.0239639),
		vec3(-34.0238266, 0.0, 188.885605),
		vec3(-61.4661980, 0.0, 124.106308),
		vec3(-10.8928108, 0.0, 129.467896),
		vec3(-45.1460953, 0.0, 61.7927818),
		vec3(-63.3019028, 0.0, 29.8712196),
		vec3(-66.2908478, 0.0, -11.7020378),
		vec3(-98.6919708, 0.0, 48.2654800),
		vec3(-32.6012917, 0.0, -40.1787720),
		vec3(-16.0055981, 0.0, -1.48840332),
		vec3(-97.6952896, 0.0, 6.50359726),
		vec3(29.4384251, 0.0, 3.15981674),
		vec3(77.4435883, 0.0, -6.43557739),
		vec3(106.919266, 0.0, 29.5377045),
		vec3(117.292038, 0.0, -37.3954849),
		vec3(67.8571777, 0.0, -65.5759659),
		vec3(50.1427460, 0.0, -39.7832603),
		vec3(7.95413971, 0.0, -40.7614212)
	};

	// Initialize the trees array with randomly rotated trees
	Tree tempTree;
	vec3 pos;
	for (int i = 0; i < TREE_COUNT; i++) { 
		if (i < sizeof(treepos) / sizeof(vec3))
			tempTree.Position = treepos[i];
		else {
			do {
				pos = PLANE_TREES * rand_vector();
			}
			while ((pos.x < ALTAR_SIZE * 3 && pos.x > -ALTAR_SIZE * 3) && 
				(pos.z < 1.5f * ALTAR_POS.z && pos.z > 0));
			pos.y = 0;
			tempTree.Position = pos;
		}
		tempTree.rotation = 2 * rand_angle();
		tempTree.size = (2 + (float)rand() / RAND_MAX) * ALTAR_SIZE;
		trees.push_back(tempTree);
	}
}

// Perform the fire spreading calculations based on distance
void spreadFlames() {
	if (held == HELD_MATCH && !transfix) {
		transfix = 1;
		for (unsigned int i = 0; i < flames.size(); i++) {
			transfix = transfix && flames[i].lit;
			if (glm::distance(match.source.Position, flames[i].source.Position) < FIRE_SPREAD) {
				if (match.lit || flames[i].lit) {
					if (!match.lit) { // flame[i].lit
						match.lit = 1;
						match.outsideColor = flames[i].outsideColor;
						match.source.Color = flames[i].source.Color;
						matchpos = flames.size();
						flames.push_back(match);
					}
					else {
						flames[i].lit = 1;
						flames[i].outsideColor = match.outsideColor;
						flames[i].source.Color = match.source.Color;
					}
				}
			}
		}
	}
}

// Transform the plane mesh to fill the screen
void draw_screen_plane(Shader_Program prog) {
	int nIndices = plane.useMesh(prog.getAttribute("aPos"));

	ModelTrans.pushMatrix();
		ModelTrans.scale(2.0f);
		ModelTrans.translate(vec3(-0.5f, -0.5f, 0));
		glUniformMatrix4fv(prog.getUniform("M"), 1, GL_FALSE, glm::value_ptr(ModelTrans.modelViewMatrix));
		glDrawElements(GL_TRIANGLES, nIndices, GL_UNSIGNED_INT, 0);
	ModelTrans.popMatrix();
}

void draw_plane(mat4 View, Shader_Program prog) {
	int nIndices = plane.useMesh(prog.getAttribute("aPos"), prog.getAttribute("aNor"));
	
	glUniform4f(prog.getUniform("UdColor"), 0.2f, 0.5f, 0.35f, 1.0f);

	glUniformMatrix4fv(prog.getUniform("M"), 1, GL_FALSE, glm::value_ptr(ModelTrans.modelViewMatrix));
	glUniformMatrix4fv(prog.getUniform("N"), 1, GL_FALSE, glm::value_ptr(glm::inverse(glm::transpose(View * ModelTrans.modelViewMatrix))));
	glDrawElements(GL_TRIANGLES, nIndices, GL_UNSIGNED_INT, 0);
}

void draw_bunny(mat4 View, Shader_Program prog) {
	int nIndices = bunny.useMesh(prog.getAttribute("aPos"), prog.getAttribute("aNor"));

	// If we have both torches lit, set the bunny to the average torch color
	if (flames[torch1pos].lit && flames[torch2pos].lit) 
		glUniform4f(prog.getUniform("UdColor"), 
		0.8f * (flames[torch1pos].source.Color.x + flames[torch2pos].source.Color.x) / 2.0f, 
		0.8f * (flames[torch1pos].source.Color.y + flames[torch2pos].source.Color.y) / 2.0f, 
		0.8f * (flames[torch1pos].source.Color.z + flames[torch2pos].source.Color.z) / 2.0f, 1.0f);
	else
		glUniform4f(prog.getUniform("UdColor"), 0.8f, 0.8f, 0.15f, 1.0f);

	ModelTrans.pushMatrix();
		ModelTrans.translate(glm::vec3(0, 2 * ALTAR_HEIGHT + ALTAR_SIZE / 2 - 0.1f, 0));
		ModelTrans.rotate(180.0f, vec3(0, 1, 0));
		ModelTrans.scale(ALTAR_SIZE / 2);

		glUniformMatrix4fv(prog.getUniform("M"), 1, GL_FALSE, glm::value_ptr(ModelTrans.modelViewMatrix));
		glUniformMatrix4fv(prog.getUniform("N"), 1, GL_FALSE, glm::value_ptr(glm::inverse(glm::transpose(View * ModelTrans.modelViewMatrix))));
		glDrawElements(GL_TRIANGLES, nIndices, GL_UNSIGNED_INT, 0);
	ModelTrans.popMatrix();
}

void draw_pedestal(mat4 View, Shader_Program prog) {
	int nIndices = pedestal.useMesh(prog.getAttribute("aPos"), prog.getAttribute("aNor"));
	
	glUniform4f(prog.getUniform("UdColor"), 0.4f, 0.05f, 0.15f, 1.0f);

	ModelTrans.pushMatrix();
		ModelTrans.translate(glm::vec3(0, ALTAR_HEIGHT, 0));
		ModelTrans.scale(ALTAR_SIZE, ALTAR_HEIGHT, ALTAR_SIZE);

		glUniformMatrix4fv(prog.getUniform("M"), 1, GL_FALSE, glm::value_ptr(ModelTrans.modelViewMatrix));
		glUniformMatrix4fv(prog.getUniform("N"), 1, GL_FALSE, glm::value_ptr(glm::inverse(glm::transpose(View * ModelTrans.modelViewMatrix))));
		glDrawElements(GL_TRIANGLES, nIndices, GL_UNSIGNED_INT, 0);
	ModelTrans.popMatrix();
}

void draw_shrine(mat4 View, Shader_Program prog) {
	int nIndices = shrine.useMesh(prog.getAttribute("aPos"), prog.getAttribute("aNor"));
	
	glUniform4f(prog.getUniform("UdColor"), 0.7f, 0.25f, 0.35f, 1.0f);

	ModelTrans.pushMatrix();
		ModelTrans.scale(ALTAR_SIZE * 3);
		ModelTrans.translate(glm::vec3(0, 0.8, 0));

		glUniformMatrix4fv(prog.getUniform("M"), 1, GL_FALSE, glm::value_ptr(ModelTrans.modelViewMatrix));
		glUniformMatrix4fv(prog.getUniform("N"), 1, GL_FALSE, glm::value_ptr(glm::inverse(glm::transpose(View * ModelTrans.modelViewMatrix))));
		glDrawElements(GL_TRIANGLES, nIndices, GL_UNSIGNED_INT, 0);
	ModelTrans.popMatrix();
}

void draw_lamp(mat4 View, Shader_Program prog) {
	int nIndices = lamp.useMesh(prog.getAttribute("aPos"), prog.getAttribute("aNor"));
	
	glUniform4f(prog.getUniform("UdColor"), 0.25f, 0.25f, 0.35f, 1.0f);

	ModelTrans.pushMatrix();
		ModelTrans.scale(2 * ALTAR_HEIGHT);
		ModelTrans.translate(glm::vec3(0, 1, 0));

		glUniformMatrix4fv(prog.getUniform("M"), 1, GL_FALSE, glm::value_ptr(ModelTrans.modelViewMatrix));
		glUniformMatrix4fv(prog.getUniform("N"), 1, GL_FALSE, glm::value_ptr(glm::inverse(glm::transpose(View * ModelTrans.modelViewMatrix))));
		glDrawElements(GL_TRIANGLES, nIndices, GL_UNSIGNED_INT, 0);
	ModelTrans.popMatrix();
}

void draw_candles(mat4 View, Shader_Program prog) {
	int nIndices = candle.useMesh(prog.getAttribute("aPos"), prog.getAttribute("aNor"));
	
	glUniform4f(prog.getUniform("UdColor"), 0.8f, 0.8f, 0.75f, 1.0f);

	for (unsigned int i = 0; i < flames.size(); i++) {
		if (i != matchpos || !match.lit) {
			ModelTrans.pushMatrix();
				ModelTrans.translate(glm::vec3(flames[i].source.Position.x, flames[i].source.Position.y / 2 + ALTAR_HEIGHT, flames[i].source.Position.z));
				ModelTrans.scale(CANDLE_SIZE, (flames[i].source.Position.y - 2 * ALTAR_HEIGHT) / 2, CANDLE_SIZE);

				glUniformMatrix4fv(prog.getUniform("M"), 1, GL_FALSE, glm::value_ptr(ModelTrans.modelViewMatrix));
				glUniformMatrix4fv(prog.getUniform("N"), 1, GL_FALSE, glm::value_ptr(glm::inverse(glm::transpose(View * ModelTrans.modelViewMatrix))));
				glDrawElements(GL_TRIANGLES, nIndices, GL_UNSIGNED_INT, 0);
			ModelTrans.popMatrix();
		}
	}
}

void draw_trees(mat4 View, Shader_Program prog) {
	int nIndices = tree.useMesh(prog.getAttribute("aPos"), prog.getAttribute("aNor"));
	glUniform4f(prog.getUniform("UdColor"), 0.38f, 0.25f, 0.08f, 1.0f);

	for (unsigned int i = 0; i < trees.size(); i++) {
		ModelTrans.pushMatrix();
			ModelTrans.translate(trees[i].Position);
			ModelTrans.rotate(trees[i].rotation, vec3(0, 1, 0));
			ModelTrans.scale(trees[i].size);
			ModelTrans.translate(glm::vec3(0, 0.9, 0));

			glUniformMatrix4fv(prog.getUniform("M"), 1, GL_FALSE, glm::value_ptr(ModelTrans.modelViewMatrix));
			glUniformMatrix4fv(prog.getUniform("N"), 1, GL_FALSE, glm::value_ptr(glm::inverse(glm::transpose(View * ModelTrans.modelViewMatrix))));
			glDrawElements(GL_TRIANGLES, nIndices, GL_UNSIGNED_INT, 0);
		ModelTrans.popMatrix();
	}
}

void draw_match(mat4 View, Shader_Program prog) {
	int nIndices;

	ModelTrans.pushMatrix();
		ModelTrans.translate(vec3(0, 0, -3.0f));
		ModelTrans.rotate(30.0f, vec3(0, 0, 1));

		// Matchstick
		nIndices = cube.useMesh(prog.getAttribute("aPos"), prog.getAttribute("aNor"));
		glUniform4f(prog.getUniform("UdColor"), 0.98f, 0.85f, 0.55f, 1.0f);

		ModelTrans.pushMatrix();
			ModelTrans.scale(0.01f, 0.07f, 0.01f);

			glUniformMatrix4fv(prog.getUniform("M"), 1, GL_FALSE, glm::value_ptr(glm::inverse(View) * ModelTrans.modelViewMatrix));
			glUniformMatrix4fv(prog.getUniform("N"), 1, GL_FALSE, glm::value_ptr(glm::inverse(glm::transpose(ModelTrans.modelViewMatrix))));
			glDrawElements(GL_TRIANGLES, nIndices, GL_UNSIGNED_INT, 0);
		ModelTrans.popMatrix();

		// Matchhead
		nIndices = sphere.useMesh(prog.getAttribute("aPos"), prog.getAttribute("aNor"));
		glUniform4f(prog.getUniform("UdColor"), 0.68f, 0.05f, 0.15f, 1.0f);

		ModelTrans.pushMatrix();
			ModelTrans.translate(vec3(0, 0.07f, 0));
			ModelTrans.scale(0.015f, 0.03f, 0.015f);

			glUniformMatrix4fv(prog.getUniform("M"), 1, GL_FALSE, glm::value_ptr(glm::inverse(View) * ModelTrans.modelViewMatrix));
			glUniformMatrix4fv(prog.getUniform("N"), 1, GL_FALSE, glm::value_ptr(glm::inverse(glm::transpose(ModelTrans.modelViewMatrix))));
			glDrawElements(GL_TRIANGLES, nIndices, GL_UNSIGNED_INT, 0);
		ModelTrans.popMatrix();
	ModelTrans.popMatrix();
}

// Perform first a horizontal, then vertical Guassian Blur
// Apply DOF, use the variable blur radius
void blurPass(GLuint texID_in, int texOffset_out, int useDOF) {
	//Horizontal
	glBindFramebuffer(GL_DRAW_FRAMEBUFFER, frameBufObj);
	glDrawBuffer(GL_COLOR_ATTACHMENT0 + TEXTURE_BLUR_H);
	glUseProgram(blurHoriz.getProgramID());
	glActiveTexture(GL_TEXTURE0);

	glBindTexture(GL_TEXTURE_2D, texID_in);

	glUniform1i(blurHoriz.getUniform("image"), 0);
	glUniform2f(blurHoriz.getUniform("uSize"), g_width, g_height);
	glUniform1f(blurHoriz.getUniform("radius"), blur_rad);
	glUniform1i(blurHoriz.getUniform("dof"), useDOF);
	glUniform1f(blurHoriz.getUniform("zFar"), ZFAR);

	glUniformMatrix4fv(blurHoriz.getUniform("V"), 1, GL_FALSE, glm::value_ptr(mat4(1.0f)));
	glUniformMatrix4fv(blurHoriz.getUniform("P"), 1, GL_FALSE, glm::value_ptr(mat4(1.0f)));

	if (dof) {
		glActiveTexture(GL_TEXTURE1);
		glBindTexture(GL_TEXTURE_2D, Tex[GBUFFER_TEXTURE_TYPE_POSITION]);
		glUniform1i(blurHoriz.getUniform("depth"), 1);
	}

	draw_screen_plane(blurHoriz);


	//Vertical
	glDrawBuffer(GL_COLOR_ATTACHMENT0 + texOffset_out);

	glUseProgram(blurVert.getProgramID());
	glActiveTexture(GL_TEXTURE0);

	glBindTexture(GL_TEXTURE_2D, blurHTex);

	glUniform1i(blurVert.getUniform("image"), 0);
	glUniform2f(blurVert.getUniform("uSize"), g_width, g_height);
	glUniform1f(blurVert.getUniform("radius"), blur_rad);
	glUniform1i(blurVert.getUniform("dof"), useDOF);
	glUniform1f(blurHoriz.getUniform("zFar"), ZFAR); 

	glUniformMatrix4fv(blurVert.getUniform("V"), 1, GL_FALSE, glm::value_ptr(mat4(1.0f)));
	glUniformMatrix4fv(blurVert.getUniform("P"), 1, GL_FALSE, glm::value_ptr(mat4(1.0f)));

	if (dof) {
		glActiveTexture(GL_TEXTURE1);
		glBindTexture(GL_TEXTURE_2D, Tex[GBUFFER_TEXTURE_TYPE_POSITION]);
		glUniform1i(blurVert.getUniform("depth"), 1);
	}

	draw_screen_plane(blurVert);
}

// Prepare the framebuffer and openGL state for the geometry pass
void geometryInit(mat4 View, mat4 Projection) {
	glUseProgram(geomPass.getProgramID());

	glBindFramebuffer(GL_DRAW_FRAMEBUFFER, frameBufObj);
	glDrawBuffer(GL_COLOR_ATTACHMENT0 + TEXTURE_FINAL);
	glClear(GL_COLOR_BUFFER_BIT);

	glBindFramebuffer(GL_DRAW_FRAMEBUFFER, frameBufObj);
	GLenum DrawBuffers[] = { GL_COLOR_ATTACHMENT0 + GBUFFER_TEXTURE_TYPE_POSITION, 
						     GL_COLOR_ATTACHMENT0 + GBUFFER_TEXTURE_TYPE_NORMAL,
						     GL_COLOR_ATTACHMENT0 + GBUFFER_TEXTURE_TYPE_DIFFUSE };
    glDrawBuffers(3, DrawBuffers);

	glDepthMask(GL_TRUE);

	// Clear the screen
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glEnable(GL_DEPTH_TEST);
	
	glUniformMatrix4fv(geomPass.getUniform("P"), 1, GL_FALSE, glm::value_ptr(Projection));
	glUniformMatrix4fv(geomPass.getUniform("V"), 1, GL_FALSE, glm::value_ptr(View));
}

// Draw all scene geometry to the G-Buffer
void geometryPass(mat4 View, mat4 Projection) {
	geometryInit(View, Projection);
	
	ModelTrans.pushMatrix();
		ModelTrans.translate(vec3(0, FLOOR, 0));
		ModelTrans.scale(2.0f * PLANE_INFINITY);
		ModelTrans.rotate(-90.0f, vec3(1, 0, 0));
		ModelTrans.translate(vec3(-0.5f, -0.5f, 0));
		draw_plane(View, geomPass);
	ModelTrans.popMatrix();

	ModelTrans.pushMatrix();
		ModelTrans.translate(ALTAR_POS);	

		draw_bunny(View, geomPass);		
		draw_pedestal(View, geomPass);
		draw_shrine(View, geomPass);
		
		ModelTrans.pushMatrix();
			ModelTrans.translate(vec3(2 * ALTAR_SIZE, 0, -2 * ALTAR_SIZE));
			draw_lamp(View, geomPass);
		ModelTrans.popMatrix();
		ModelTrans.pushMatrix();
			ModelTrans.translate(vec3(-2 * ALTAR_SIZE, 0, -2 * ALTAR_SIZE));
			draw_lamp(View, geomPass);
		ModelTrans.popMatrix();
	
		draw_candles(View, geomPass);
	ModelTrans.popMatrix();

	draw_trees(View, geomPass);

	if (held == HELD_MATCH) {
		draw_match(View, geomPass);

		// Position is relative to the eye (positioned in viewspace)
		match.source.Position = vec3(glm::inverse(View) * vec4(-0.02f, 0.01f, -3.0f, 1.0f)) - ALTAR_POS;
		if (match.lit) 
			flames[matchpos].source.Position = match.source.Position;
	}

	// Disable and unbind
	GLSL::disableVertexAttribArray(geomPass.getAttribute("aPos"));
	GLSL::disableVertexAttribArray(geomPass.getAttribute("aNor"));
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);

	glDepthMask(GL_FALSE);	
}

// Generate the ambient occlusion texture based on scene geometry
void ambientOcclusionPass() {
	glDrawBuffer(GL_COLOR_ATTACHMENT0 + TEXTURE_AMB);
	glClear(GL_COLOR_BUFFER_BIT);

	if (amb && (drawMode == DRAW_FULL || drawMode == DRAW_AMB)) {
		glDisable(GL_DEPTH_TEST);
		glUseProgram(ambO.getProgramID());

		glActiveTexture(GL_TEXTURE0 + GBUFFER_TEXTURE_TYPE_POSITION);
		glBindTexture(GL_TEXTURE_2D, Tex[GBUFFER_TEXTURE_TYPE_POSITION]);
		glUniform1i(ambO.getUniform("posMap"), GBUFFER_TEXTURE_TYPE_POSITION);

		glActiveTexture(GL_TEXTURE0 + GBUFFER_TEXTURE_TYPE_NORMAL);
		glBindTexture(GL_TEXTURE_2D, Tex[GBUFFER_TEXTURE_TYPE_NORMAL]);
		glUniform1i(ambO.getUniform("norMap"), GBUFFER_TEXTURE_TYPE_NORMAL);

		glUniform2f(ambO.getUniform("filterRadius"), ambO_rad / g_width, ambO_rad / g_height);
		glUniform1f(ambO.getUniform("distanceThreshold"), ambO_dist);

		draw_screen_plane(ambO);

		if (blur)
		blurPass(ambTex, TEXTURE_AMB, 0);
	}
}

// Prepare for the lighting passes (stencil and point lights)
int lightingInit(mat4 View, mat4 Projection) {
	//Stencil Pass
	glUseProgram(nullPass.getProgramID());
	glUniformMatrix4fv(nullPass.getUniform("V"), 1, GL_FALSE, glm::value_ptr(View));
	glUniformMatrix4fv(nullPass.getUniform("P"), 1, GL_FALSE, glm::value_ptr(Projection));
	
	int nIndices = sphere.useMesh(nullPass.getAttribute("aPos"));

	//Light Pass
	vec3 eye = cam.getEye();
	glUseProgram(lightPass.getProgramID());
	glUniform1i(lightPass.getUniform("posMap"), GBUFFER_TEXTURE_TYPE_POSITION);
	glUniform1i(lightPass.getUniform("colMap"), GBUFFER_TEXTURE_TYPE_DIFFUSE);
	glUniform1i(lightPass.getUniform("norMap"), GBUFFER_TEXTURE_TYPE_NORMAL);
	glUniform1i(lightPass.getUniform("ambMap"), 3);
	glUniform3f(lightPass.getUniform("uEye"), eye.x, eye.y, eye.z);
	glUniform2f(lightPass.getUniform("uSize"), (float)g_width, (float)g_height);
	glUniformMatrix4fv(lightPass.getUniform("V"), 1, GL_FALSE, glm::value_ptr(View));
	glUniformMatrix4fv(lightPass.getUniform("P"), 1, GL_FALSE, glm::value_ptr(Projection));
	
	nIndices = sphere.useMesh(lightPass.getAttribute("aPos"));

	return nIndices;
}

// Per-light preparation for the stencil pass
void lighting_stencilSetup() {
	glUseProgram(nullPass.getProgramID());
	glEnable(GL_STENCIL_TEST);

	glDrawBuffer(GL_NONE);
	glEnable(GL_DEPTH_TEST);
	glDisable(GL_CULL_FACE);

	glClear(GL_STENCIL_BUFFER_BIT);
	glStencilFunc(GL_ALWAYS, 0, 0);
	glStencilOpSeparate(GL_BACK, GL_KEEP, GL_INCR_WRAP, GL_KEEP);
	glStencilOpSeparate(GL_FRONT, GL_KEEP, GL_DECR_WRAP, GL_KEEP);
}

// Per-light stencil fill to ensure only relevant areas are lit
void lighting_stencilPass(int flameDex, int nIndices) {
	ModelTrans.pushMatrix();
		ModelTrans.translate(ALTAR_POS);
		ModelTrans.translate(flames[flameDex].source.Position + vec3(0, CANDLE_WICK, 0));
		ModelTrans.scale(calcLightSphere(source));

		glUniformMatrix4fv(nullPass.getUniform("M"), 1, GL_FALSE, glm::value_ptr(ModelTrans.modelViewMatrix));
		glDrawElements(GL_TRIANGLES, nIndices, GL_UNSIGNED_INT, 0);
	ModelTrans.popMatrix();
}

// Per-light preparation for the point lighting pass
void lighting_pointLightSetup() {
	glDrawBuffer(GL_COLOR_ATTACHMENT0 + TEXTURE_FINAL);
	glUseProgram(lightPass.getProgramID());
	for (int j = 0; j < GBUFFER_NUM_TEXTURES; j++) {
		glActiveTexture(GL_TEXTURE0 + j);
		glBindTexture(GL_TEXTURE_2D, Tex[j]);
	}
	glActiveTexture(GL_TEXTURE0 + 3);
	glBindTexture(GL_TEXTURE_2D, ambTex);

	glStencilFunc(GL_NOTEQUAL, 0, 0xFF);
	glDisable(GL_DEPTH_TEST);

	glEnable(GL_BLEND);
	glBlendEquation(GL_FUNC_ADD);
	glBlendFunc(GL_ONE, GL_ONE);

	glEnable(GL_CULL_FACE);
	glCullFace(GL_FRONT);
}

// Per-light execution of Phong Lighting within the stenciled area
void lighting_pointLightPass(int flameDex, int nIndices) {
	ModelTrans.pushMatrix();
		ModelTrans.translate(ALTAR_POS);
		ModelTrans.translate(flames[flameDex].source.Position + vec3(0, CANDLE_WICK, 0));
		ModelTrans.scale(calcLightSphere(source));

		vec4 lightPos = vec4(flames[flameDex].source.Position.x, flames[flameDex].source.Position.y + CANDLE_WICK, flames[flameDex].source.Position.z, 1.0);
		lightPos = glm::translate(ALTAR_POS) * lightPos;
		glUniform3f(lightPass.getUniform("uLightCol"), flames[flameDex].source.Color.r, flames[flameDex].source.Color.g, flames[flameDex].source.Color.b);
		glUniform3f(lightPass.getUniform("uLightAtt"), flames[flameDex].source.Attenuation.x, flames[flameDex].source.Attenuation.y, flames[flameDex].source.Attenuation.z);
		glUniform1f(lightPass.getUniform("uAmbIntensity"), flames[flameDex].source.AmbIntensity);
		glUniform1f(lightPass.getUniform("uDifIntensity"), flames[flameDex].source.DifIntensity);
		glUniform1f(lightPass.getUniform("uSpeIntensity"), flames[flameDex].source.SpeIntensity);
		glUniform1f(lightPass.getUniform("uSpePower"), flames[flameDex].source.SpePower);
		glUniform3f(lightPass.getUniform("uLightPos"), lightPos.x, lightPos.y, lightPos.z);

		glUniformMatrix4fv(lightPass.getUniform("M"), 1, GL_FALSE, glm::value_ptr(ModelTrans.modelViewMatrix));
		glDrawElements(GL_TRIANGLES, nIndices, GL_UNSIGNED_INT, 0);
	ModelTrans.popMatrix();

	glCullFace(GL_BACK);
	glDisable(GL_BLEND);
}

// Prepare the openGL state for the fullscreen directional light
void lighting_dirLightSetup(mat4 View, mat4 Projection) {
	glUseProgram(dirLightPass.getProgramID());
		glUniform1i(dirLightPass.getUniform("posMap"), GBUFFER_TEXTURE_TYPE_POSITION);
		glUniform1i(dirLightPass.getUniform("colMap"), GBUFFER_TEXTURE_TYPE_DIFFUSE);
		glUniform1i(dirLightPass.getUniform("norMap"), GBUFFER_TEXTURE_TYPE_NORMAL);
		glUniform1i(dirLightPass.getUniform("ambMap"), 3);
		vec4 eye_vspace = View * vec4(cam.getEye(), 1.0f);
		glUniform3f(dirLightPass.getUniform("uEye"), eye_vspace.x, eye_vspace.y, eye_vspace.z);
		glUniform2f(dirLightPass.getUniform("uSize"), (float)g_width, (float)g_height);
	
		glUniform3f(dirLightPass.getUniform("uLightCol"), moon.Color.r, moon.Color.g, moon.Color.b);
		glUniform3f(dirLightPass.getUniform("uLightDirection"), moon.Position.x, moon.Position.y, moon.Position.z);
		glUniform1f(dirLightPass.getUniform("uAmbIntensity"), moon.AmbIntensity);
		glUniform1f(dirLightPass.getUniform("uDifIntensity"), moon.DifIntensity);
		glUniform1f(dirLightPass.getUniform("uSpeIntensity"), moon.SpeIntensity);
		glUniform1f(dirLightPass.getUniform("uSpePower"), moon.SpePower);

		glUniformMatrix4fv(dirLightPass.getUniform("V"), 1, GL_FALSE, glm::value_ptr(mat4(1.0f)));
		glUniformMatrix4fv(dirLightPass.getUniform("P"), 1, GL_FALSE, glm::value_ptr(mat4(1.0f)));

		glDrawBuffer(GL_COLOR_ATTACHMENT0 + TEXTURE_FINAL);
		glUseProgram(dirLightPass.getProgramID());
		for (int j = 0; j < GBUFFER_NUM_TEXTURES; j++) {
			glActiveTexture(GL_TEXTURE0 + j);
			glBindTexture(GL_TEXTURE_2D, Tex[j]);
		}
		glActiveTexture(GL_TEXTURE0 + 3);
		glBindTexture(GL_TEXTURE_2D, ambTex);

		glDisable(GL_DEPTH_TEST);
		glEnable(GL_BLEND);
		glBlendEquation(GL_FUNC_ADD);
		glBlendFunc(GL_ONE, GL_ONE);
}

// Perform all the lighting in the scene (point lights, directional light)
void lightingPass(mat4 View, mat4 Projection) {
	if (drawMode == DRAW_FULL) {
		int nIndices = lightingInit(View, Projection);

		for (unsigned int i = 0; i < flames.size(); i++) {
			if (flames[i].lit) {
				lighting_stencilSetup();
				lighting_stencilPass(i, nIndices);

				lighting_pointLightSetup();
				lighting_pointLightPass(i, nIndices);
			}
		}
		glDisable(GL_STENCIL_TEST);
		
		lighting_dirLightSetup(View, Projection);
		draw_screen_plane(dirLightPass);

		glDisable(GL_BLEND);
	}
	glDisable(GL_DEPTH_TEST);
}

// Draw the skybox, depth-tested to fill in empty spaces in the distance
void skyboxPass(mat4 View, mat4 Projection) {
	glUseProgram(skyDraw.getProgramID());
	glDrawBuffer(GL_COLOR_ATTACHMENT0 + TEXTURE_FINAL);

	glEnable(GL_DEPTH_TEST);
	glDisable(GL_CULL_FACE);

	glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_CUBE_MAP, skyboxTex);

	int nIndices = sphere.useMesh(skyDraw.getAttribute("aPos")); // Sphere shows less seams than cube?
	glUniform1i(skyDraw.getUniform("cubeTex"), 0);
	glUniform3f(skyDraw.getUniform("lightColor"), moon.Color.r / 3.0f, moon.Color.g / 3.0f, moon.Color.b / 3.0f);
	glUniformMatrix4fv(skyDraw.getUniform("P"), 1, GL_FALSE, glm::value_ptr(Projection));
	glUniformMatrix4fv(skyDraw.getUniform("V"), 1, GL_FALSE, glm::value_ptr(View));

	ModelTrans.pushMatrix();
		ModelTrans.translate(cam.getEye());
		ModelTrans.scale(270.0f); // Arbitrary limit within 300.0f zFar
		ModelTrans.rotate(45.0f, vec3(0, 1, 0));
		glUniformMatrix4fv(skyDraw.getUniform("M"), 1, GL_FALSE, glm::value_ptr(ModelTrans.modelViewMatrix));
		glDrawElements(GL_TRIANGLES, nIndices, GL_UNSIGNED_INT, 0);
	ModelTrans.popMatrix();

	glEnable(GL_CULL_FACE);
	glDisable(GL_DEPTH_TEST);
}

// Loop through and draw all components of each flame, and update their positions
void firePass(mat4 View, mat4 Projection) {
	glUseProgram(fireDraw.getProgramID());
	glDrawBuffer(GL_COLOR_ATTACHMENT0 + TEXTURE_FIRE);
	//glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
	glClear(GL_COLOR_BUFFER_BIT);
	glClearColor(0.0f, 0.0f, 0.0f, 1.0f);

	glEnable(GL_DEPTH_TEST);
	glEnable(GL_BLEND);

	glBlendEquation(GL_FUNC_ADD);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	glEnable(GL_CULL_FACE);
	glCullFace(GL_BACK);

	int nIndices = fire.useMesh(fireDraw.getAttribute("aPos"), fireDraw.getAttribute("aNor"));
	glUniformMatrix4fv(fireDraw.getUniform("P"), 1, GL_FALSE, glm::value_ptr(Projection));
	glUniformMatrix4fv(fireDraw.getUniform("V"), 1, GL_FALSE, glm::value_ptr(View));

	ModelTrans.pushMatrix();
		ModelTrans.translate(ALTAR_POS);
		for (unsigned int i = 0; i < flames.size(); i++) {
			if (flames[i].lit) {
				ModelTrans.pushMatrix();
					ModelTrans.translate(vec3(flames[i].source.Position.x, flames[i].source.Position.y + CANDLE_WICK / 2, flames[i].source.Position.z));

					glUniform3f(fireDraw.getUniform("colCenter"), flames[i].source.Color.r, flames[i].source.Color.g, flames[i].source.Color.b);
					glUniform3f(fireDraw.getUniform("colOutside"), flames[i].outsideColor.r, flames[i].outsideColor.g, flames[i].outsideColor.b);
					glUniform1f(fireDraw.getUniform("alphaFactor"), 0.3f);

					for (int j = 0; j < FLAME_PARTICLE_COUNT; j++) {
						ModelTrans.pushMatrix();
							ModelTrans.translate(vec3(0, CANDLE_WICK * flames[i].life[j], 0) + flames[i].sway[j]);
							ModelTrans.scale(CANDLE_WICK);
							ModelTrans.scale(0.8f * (1.1f - flames[i].life[j]), flames[i].life[j] + 0.2f, 0.8f * (1.1f - flames[i].life[j]));

							glUniformMatrix4fv(fireDraw.getUniform("M"), 1, GL_FALSE, glm::value_ptr(ModelTrans.modelViewMatrix));
							glUniformMatrix4fv(fireDraw.getUniform("N"), 1, GL_FALSE, glm::value_ptr(glm::inverse(glm::transpose(View * ModelTrans.modelViewMatrix))));
							glDrawElements(GL_TRIANGLES, nIndices, GL_UNSIGNED_INT, 0);
						ModelTrans.popMatrix();
				
						// Update Fire
						flames[i].life[j] += 0.01f;
						if (flames[i].life[j] > 1.0f) {
							flames[i].life[j] = 0.0f;
						}

						flames[i].sway[j] += rand_vector() * 0.001f;
						if (flames[i].sway[j].x > MAX_FIRE_SWAY || flames[i].sway[j].x < -MAX_FIRE_SWAY) {
							flames[i].sway[j].x = 0.0f;
						}
						if (flames[i].sway[j].y > MAX_FIRE_SWAY || flames[i].sway[j].y < -MAX_FIRE_SWAY) {
							flames[i].sway[j].y = 0.0f;
						}
						if (flames[i].sway[j].z > MAX_FIRE_SWAY || flames[i].sway[j].z < -MAX_FIRE_SWAY) {
							flames[i].sway[j].z = 0.0f;
						}
					}
				ModelTrans.popMatrix();
			}
		}
	ModelTrans.popMatrix();

	glDisable(GL_BLEND);
	glDisable(GL_DEPTH_TEST);

	// Blur the flames appropriately
	float swap = blur_rad;
	blur_rad *= ALTAR_SIZE / glm::distance(cam.getEye(), ALTAR_POS); // Base blur radius on the distance from the altar (better results for flames in the distance)
	for (int i = 0; i < blur; i++)
		blurPass(fireTex, TEXTURE_FIRE, 0);
	blur_rad = swap;
}

// Output the full scene to the screen
void drawFinal() {
	glBindFramebuffer(GL_DRAW_FRAMEBUFFER, 0);
	glBindFramebuffer(GL_READ_FRAMEBUFFER, frameBufObj);
		
	glUseProgram(texDraw.getProgramID());
	glActiveTexture(GL_TEXTURE0);
	
	glUniform1i(texDraw.getUniform("drawDepth"), 0);

	// Choose the drawn texture
	if (drawMode == DRAW_FULL)
		glBindTexture(GL_TEXTURE_2D, finalTex);
	else if (drawMode == DRAW_DEPTH) {
		glBindTexture(GL_TEXTURE_2D, depthTex);
		glUniform1i(texDraw.getUniform("drawDepth"), 1);
		glUniform1f(texDraw.getUniform("zNear"), ZNEAR);
		glUniform1f(texDraw.getUniform("zFar"), ZFAR);
	}
	else if (drawMode == DRAW_AMB)
		glBindTexture(GL_TEXTURE_2D, ambTex);
	else
		glBindTexture(GL_TEXTURE_2D, Tex[drawMode - 1]);

	glUniform1i(texDraw.getUniform("texture0"), 0);
	draw_screen_plane(texDraw);
	
	// Add Fire
	if (drawMode == DRAW_FULL) {
		glEnable(GL_BLEND);
		glBlendEquation(GL_FUNC_ADD);
		glBlendFunc(GL_ONE, GL_ONE);
		
		glBindTexture(GL_TEXTURE_2D, fireTex);
		draw_screen_plane(texDraw);

		glDisable(GL_BLEND);
	}

	// Add bloom component on top, if enabled
	if (bloom && blur && drawMode == DRAW_FULL) {
		glUseProgram(bloomPass.getProgramID());
		glEnable(GL_BLEND);
		glBlendEquation(GL_FUNC_ADD);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE);

		glBindTexture(GL_TEXTURE_2D, blurVTex);
		glUniform1i(bloomPass.getUniform("texture0"), 0);
		draw_screen_plane(bloomPass);

		glDisable(GL_BLEND);
	}
}

// Perform full per-frame drawing in the pipeline
void drawGL()
{
	glm::mat4 Projection = glm::perspective(45.0f, (float)g_width/g_height, ZNEAR, ZFAR); 
	glm::mat4 View;
	if (transfix)
		View = cam.getViewLookat(ALTAR_POS + vec3(0, 2 * ALTAR_HEIGHT, 0));
	else
		View = cam.getView();

	ModelTrans.loadIdentity();

	// Process the actual content and lighting across multiple passes
	geometryPass(View, Projection);
	ambientOcclusionPass();
	lightingPass(View, Projection);
	skyboxPass(View, Projection);
	firePass(View, Projection);

	// Postprocessing Blur Pass (DOF, BLOOM)
	if ((blur || bloom) && drawMode == DRAW_FULL) {
		for (int blurcnt = 0; blurcnt < blur; blurcnt++) {
			blurPass((blurcnt) ? blurVTex : finalTex, // Only read from "final tex" on the first pass, else blur_v
				(bloom || blurcnt + 1 < blur) ? TEXTURE_BLUR_V : TEXTURE_FINAL, // Write to blur_v on all but the last (or the bloom) pass
				dof);
		}
	}
	
	// Output the chosen texture to the screen based on the current mode
	drawFinal();
	
	glUseProgram(0);
	
	assert(glGetError() == GL_NO_ERROR);
}

int main(int argc, char **argv)
{
	// Initialize globals
	g_width = 1024;
	g_height = 768;
	drawMode = DRAW_FULL;
	blur = 1;
	bloom = 0;
	amb = 1;
	dof = 1;
	transfix = 0;
	held = HELD_NONE;
	bright = 0;

	setupGLFW();

	// Initialize glad
	if (glewInit() != GLEW_OK) {
		fprintf(stderr, "Unable to initialize glad");
		glfwDestroyWindow(window);
		glfwTerminate();
		return 1;
	}

	initGL();

	loadMeshes();
	loadShaders();
	loadSkybox();

	setupLights();

	generate_scene();

	int frames = 0;
	double thisFrame, lastFrame = 0.0;
	glfwSetTime(0);
	do{
		thisFrame = glfwGetTime();
		// Basic frame time locking and frame-independent speeds
		if (frames < thisFrame * MAX_FPS) {
			//printf("%2.1lf\n", 1 / (thisFrame - lastFrame));
			drawGL();
			frames++;

			// Swap buffers
			glfwSwapBuffers(window);
			glfwPollEvents();
			
			spreadFlames();

			// Process keyboard-controlled movement
			vec3 lookat;
			if (transfix) // Stuck looking at the altar or not
				lookat = glm::normalize(ALTAR_POS + vec3(0, 2 * ALTAR_HEIGHT, 0) - cam.getEye());
			else
				lookat = cam.getLookat();
			cam.translate(window, lookat, (float)(thisFrame - lastFrame));

			lastFrame = thisFrame;
		}
	} // Check if the ESC key was pressed or the window was closed
	while( glfwGetKey(window, GLFW_KEY_ESCAPE ) != GLFW_PRESS &&
			glfwWindowShouldClose(window) == 0 );

	// Close OpenGL window and terminate GLFW
	glfwTerminate();

	return 0;
}
