// Michael Granneman

#include <vector>
#include "shader_program.h"
#include "load_shader.hpp"

#define UNSET -1

Shader_Program::Shader_Program() {
	vertexshader = fragshader = programID = UNSET;
}

// Compile both passed in shaders into a full program
Shader_Program::Shader_Program(string fragshader, string vertexshader) {
	setFragShader(fragshader);
	setVertexShader(vertexshader);
	loadProgram();

	queryAttributes();
	queryUniforms();
}

void Shader_Program::queryAttributes() {
	int total = -1;
	glGetProgramiv(programID, GL_ACTIVE_ATTRIBUTES, &total ); 
	for(int i=0; i<total; ++i)  {
		int name_len=-1, num=-1;
		GLenum type = GL_ZERO;
		char name[100];
		glGetActiveAttrib(programID, GLuint(i), sizeof(name)-1,
			&name_len, &num, &type, name);
		name[name_len] = 0;
		addAttribute(name);
	}
}

void Shader_Program::queryUniforms() {
	int total = -1;
	glGetProgramiv(programID, GL_ACTIVE_UNIFORMS, &total ); 
	for(int i=0; i<total; ++i)  {
		int name_len=-1, num=-1;
		GLenum type = GL_ZERO;
		char name[100];
		glGetActiveUniform(programID, GLuint(i), sizeof(name)-1,
			&name_len, &num, &type, name );
		name[name_len] = 0;
		addUniform(name);
	}
}

GLuint Shader_Program::addAttribute(string name) {
	GLuint id;

	if (compiled()) {
		id = glGetAttribLocation(programID, name.c_str());
		attribute[name] = id;
	}

	return id;
}

GLuint Shader_Program::addUniform(string name) {
	GLuint id;

	if (compiled()) {
		id = glGetUniformLocation(programID, name.c_str());
		uniform[name] = id;
	}

	return id;
}

void Shader_Program::setFragShader(string name) {
	if (!compiled()) {
		fragshader = LoadShader(name, GL_FRAGMENT_SHADER);
	}
}

void Shader_Program::setVertexShader(string name) {
	if (!compiled()) {
		vertexshader = LoadShader(name, GL_VERTEX_SHADER);
	}
}

// Link existing shaders, if they are set, into a program ready for use
// ADAPTED FROM SHADER.CPP FROM OPENGL-TUTORIALS.ORG
GLuint Shader_Program::loadProgram() {
	GLint Result = GL_FALSE;
	int InfoLogLength;

	// Link the program
	printf("Linking program\n");
	programID = glCreateProgram();

	if (vertexshader != UNSET) {
		glAttachShader(programID, vertexshader);
	}

	if (fragshader != UNSET) {
		glAttachShader(programID, fragshader);
	}

	glLinkProgram(programID);

	// Check the program
	glGetProgramiv(programID, GL_LINK_STATUS, &Result);
	glGetProgramiv(programID, GL_INFO_LOG_LENGTH, &InfoLogLength);
	if ( InfoLogLength > 0 ){
		std::vector<char> ProgramErrorMessage(InfoLogLength+1);
		glGetProgramInfoLog(programID, InfoLogLength, NULL, &ProgramErrorMessage[0]);
		printf("%s\n", &ProgramErrorMessage[0]);
	}

	if (vertexshader != UNSET) {
		glDeleteShader(vertexshader);
	}
	if (fragshader != UNSET) {
		glDeleteShader(fragshader);
	}

	return programID;
}