// Michael Granneman

#ifndef SHADER_PROGRAM_H
#define SHADER_PROGRAM_H

#include <map>
#include <string>

#include <GL/glew.h>
#include <GLFW/glfw3.h>
#include <GLM/glm.hpp>


using namespace std;

// Represents a GLSL program
// Provides easy access to attributes/uniforms by name
// Compiles program shaders and builds them
class Shader_Program {
private:
	map<string, GLuint> attribute;
	map<string, GLuint> uniform;

	GLuint fragshader;
	GLuint vertexshader;

	GLuint programID;

public:
	Shader_Program();
	Shader_Program(string fragshader, string vertexshader);

	void Shader_Program::queryAttributes();
	void Shader_Program::queryUniforms();

	GLuint addAttribute(string name);
	GLuint addUniform(string name);

	GLuint getAttribute(string name) { return attribute[name]; }
	GLuint getUniform(string name) { return uniform[name]; }

	GLuint getProgramID() { return programID; }

	void setFragShader(string name);
	void setVertexShader(string name);

	GLuint loadProgram();

	bool compiled() { return programID != -1; }
};

#endif