#version 330

in vec2 vTex;
 
uniform sampler2D norMap;
uniform sampler2D posMap;
 
uniform float distanceThreshold;
uniform vec2 filterRadius;

out float fragColor;
 
const int sample_count = 16;
const vec2 poisson16[] = vec2[](    // These are the Poisson Disk Samples
                                vec2( -0.94201624,  -0.39906216 ),
                                vec2(  0.94558609,  -0.76890725 ),
                                vec2( -0.094184101, -0.92938870 ),
                                vec2(  0.34495938,   0.29387760 ),
                                vec2( -0.91588581,   0.45771432 ),
                                vec2( -0.81544232,  -0.87912464 ),
                                vec2( -0.38277543,   0.27676845 ),
                                vec2(  0.97484398,   0.75648379 ),
                                vec2(  0.44323325,  -0.97511554 ),
                                vec2(  0.53742981,  -0.47373420 ),
                                vec2( -0.26496911,  -0.41893023 ),
                                vec2(  0.79197514,   0.19090188 ),
                                vec2( -0.24188840,   0.99706507 ),
                                vec2( -0.81409955,   0.91437590 ),
                                vec2(  0.19984126,   0.78641367 ),
                                vec2(  0.14383161,  -0.14100790 )
                               );
 
void main()
{
    vec3 viewPos = texture(posMap, vTex).xyz;
    vec3 viewNormal = normalize(texture(norMap, vTex).xyz);
 
    float ambientOcclusion = 0;
    for (int i = 0; i < sample_count; ++i)
    {
        vec2 sampleTexCoord = vTex + (poisson16[i] * (filterRadius));
        float sampleDepth = texture(posMap, sampleTexCoord).z;
        vec3 samplePos = vec3(texture(posMap, vTex).xy, sampleDepth);
        vec3 sampleDir = normalize(samplePos - viewPos);
 
        float NdotS = max(dot(viewNormal, sampleDir), 0);
        float VPdistSP = distance(viewPos, samplePos);
 
        float a = 1.0 - smoothstep(distanceThreshold, distanceThreshold * 2, VPdistSP);
        float b = NdotS;
 
        ambientOcclusion += (a * b);
    }
 
    fragColor = (ambientOcclusion / sample_count);
}