#version 330

uniform vec3 uLightCol;
uniform vec3 uLightDirection;
uniform float uAmbIntensity;
uniform float uDifIntensity;
uniform float uSpeIntensity;
uniform float uSpePower;

uniform sampler2D posMap;
uniform sampler2D norMap;
uniform sampler2D colMap;
uniform sampler2D ambMap;

uniform vec2 uSize;
uniform vec3 uEye;

vec2 CalcTexCoord()
{
    return gl_FragCoord.xy / uSize;
}

vec4 CalcLightInternal(vec3 LightDirection, vec3 WorldPos, vec3 Normal)
{
    vec4 AmbientColor = vec4(uLightCol * uAmbIntensity * (1.0 - texture(ambMap, CalcTexCoord()).r), 1.0);
    float DiffuseFactor = dot(Normal, LightDirection);

    vec4 DiffuseColor  = vec4(0, 0, 0, 0);
    vec4 SpecularColor = vec4(0, 0, 0, 0);

    if (DiffuseFactor > 0.0) {
        DiffuseColor = vec4(uLightCol, 1.0) * uDifIntensity * DiffuseFactor;

        vec3 VertexToEye = normalize(uEye - WorldPos);
		vec3 HalfDir = normalize(LightDirection + VertexToEye);
        float SpecularFactor = dot(Normal, HalfDir);
        SpecularFactor = pow(SpecularFactor, uSpePower);
        if (SpecularFactor > 0.0) {
            SpecularColor = vec4(uLightCol, 1.0) * uSpeIntensity * SpecularFactor;
        }
    }

    return (AmbientColor + DiffuseColor + SpecularColor);
}

vec4 CalcDirLight(vec3 WorldPos, vec3 Normal)
{
    vec3 LightDirection = normalize(uLightDirection);

    return CalcLightInternal(LightDirection, WorldPos, Normal);
}

out vec4 FragColor;

void main()
{
    vec2 TexCoord = CalcTexCoord();
	vec3 WorldPos = texture(posMap, TexCoord).xyz;
	vec3 Color = texture(colMap, TexCoord).xyz;
	vec3 Normal = texture(norMap, TexCoord).xyz;
	Normal = normalize(Normal);

    //FragColor = vec4(0, 1, 0, 1) + 0.0001 * vec4(Color, 1.0) + 0.0001 * CalcDirLight(WorldPos, Normal);
	FragColor = vec4(Color, 1.0) * CalcDirLight(WorldPos, Normal);
}